import time
import platform
import sqlite3
import subprocess
import re
import logging
import os

DBname = '/var/lib/kylin-system-updater/InstallDebs_TimeDB.db'
def _Utils_extract_numbers(s):
    return [float(num) for num in re.findall(r'[-+]?\d*\.\d+|\d+', s)]

# 由于没有numpy库, 手写了一个二维数组归一化方法
def normalize_by_column(matrix:list):
    # 翻转二维数组
    def rotate_by_column(matrix):
        rotated = list(zip(*matrix))
        return [list(row) for row in rotated]
    
    # 归一化一维数组
    def normalize_rows(matrix):
        return [
            # [float(element) / sum(row) for element in row]
            [float(element) / max(row) for element in row ]
            for row in matrix
        ]
    new_matrix = rotate_by_column(matrix)
    results = normalize_rows(new_matrix)
    results = rotate_by_column(results)
    return results


class calTimeManager():
    HWSPEED_DICT = {
        'x86_64':[[800, 20000, 3000, 500],[700.38, 16280.01, 3000, 327]],
        'aarch64':[[800, 20000, 3000, 500],[661.84, 17382.40, 154.02, 102.67]],
        'common':[[800, 20000, 3000, 500],[700.38, 16280.01, 3000, 327]],
    }
    def __init__(self) -> None:
        pass
    def CalInstallTimeByDebsList_min(self, debList:list) -> float:
        result = 0
        try:
            self.architecture = platform.machine()
            if self.architecture != 'x86_64' or self.architecture != 'aarch64':
                self.table_name ='debsTimeTable_common'
                self.architecture = 'common'
            else:
                self.table_name = 'debsTimeTable_' + self.architecture
            self.timeDivisorMatrix = self.HWSPEED_DICT[self.architecture]
            # Init Database
            conn = sqlite3.connect(DBname)
            table_name = 'debsTimeTable_'+self.architecture
            logging.info(f"The database is {table_name}")
            cursor = conn.cursor()

            if self._check_ifDBhasTableforarchitecture(cursor, table_name) is False:
                raise Exception(f'初始化数据库错误，缺少{self.architecture}架构的基础数据')

            # 获取基准信息
            datas, not_found_items = self._getAllInfoForDBTable(cursor, debList)
            conn.close()
            # print(datas)
            unpackTimeSum = 0
            configTimeSum = 0
            triggerTimeSum = 0
            for (name,unpackTime,configTime,triggerTime) in datas:
                unpackTimeSum += unpackTime
                configTimeSum += configTime
                triggerTimeSum += triggerTime
            
            # 根据当前硬件计算系数
            self.HwHelper = HardwareHelper()
            _localCpuSpeed = self.HwHelper._testCpuScore()
            _localMemSpeed = self.HwHelper._testMemScore()
            _localIOReadSpeed, _localIOWrittenSpeed = self.HwHelper._testIOScore()

            # 根据硬件参数和数据库采集的硬件参数，计算系数矩阵
            localSpeed = [_localCpuSpeed, _localMemSpeed, _localIOReadSpeed, _localIOWrittenSpeed]
            self.timeDivisorMatrix.append(localSpeed)
            self.timeDivisorMatrix = normalize_by_column(self.timeDivisorMatrix)
            # print(self.timeDivisorMatrix)
            
            # 计算系数
            AimHwScore = self.timeDivisorMatrix[1][0] * 0.2 + self.timeDivisorMatrix[1][1] * 0.2 + self.timeDivisorMatrix[1][2] * 0.4 + self.timeDivisorMatrix[1][3] * 0.4
            localHwScore = self.timeDivisorMatrix[2][0] * 0.2 + self.timeDivisorMatrix[2][1] * 0.2 + self.timeDivisorMatrix[2][2] * 0.4 + self.timeDivisorMatrix[2][3] * 0.4
            Index = AimHwScore/localHwScore
            # 拟合公式，10为意义上的理想硬件模拟因子，11.5为标准因子
            Index = (10 + Index) / 11.5
            # print(Index)
            result += (unpackTimeSum+configTimeSum+triggerTimeSum) * Index
            result += self._calTime(not_found_items)
            result = result / 60 + 1

        except Exception as e:
            logging.error(e)
        return int(result)

    # # 外部接口
    # def get_LastTime(self):
    #     return self.LastTime

    # # 内部接口
    # def _run(self) -> None:
    #     time.sleep(1)
    #     self.LastTime -= 1
    def _calTime(self,not_found_items):
        # 判断是否数据库中没有的包为内核包，如果是，则加上固定估计时间
        # 内核包编译安装时间大概为10分钟，经过检测
        result = 0
        isLinuxKernelDebs = False
        for items in not_found_items:
            name = re.sub(r'\d+', '', items)
            if name == "linux-headers-..--generic":
                isLinuxKernelDebs = True
                break
            elif name == "linux-headers-..-":
                isLinuxKernelDebs = True
                break
            elif name == "linux-imgae-..--generic":
                isLinuxKernelDebs = True
                break
        if isLinuxKernelDebs is True:
            result += 600
        
        # 如果还有不存在的数据，则按照50个包15分钟的时间判断
        if len(not_found_items) > 50:
            result += 15 * 60 * (len(not_found_items)/50)
        return result
            
    def _check_ifDBhasTableforarchitecture(self, cursor, table_name):
        try:
            cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
            tables = cursor.fetchall()
            logging.info(tables)
        except Exception as e:
            logging.error(e)
            return False
        if (table_name,) in tables:
            return True
        else:
            return False

    def _getAllInfoForDBTable(self, cursor, debList):
        all_data = None
        rectifyDebList = []
        for deb in debList:
            deb = deb.replace("#auto","")
            rectifyDebList.append(deb)
        try:
            placeholders = ','.join('?' for _ in rectifyDebList)
            query = f"SELECT * FROM {self.table_name} WHERE package_name IN ({placeholders})"
            # Execute the query with the list of package names
            cursor.execute(query, rectifyDebList)
            all_data = cursor.fetchall()
            # 提取找到的值
            found_items = [row[0] for row in all_data]
            # 列表中未找到的行
            not_found_items = list(set(rectifyDebList) - set(found_items))
            logging.info(f"Packages that are missing from the database:  {not_found_items}")
            logging.info(f"The quantity that is missing from the database:{len(not_found_items)}")
        except Exception as e:
            logging.error(e)
            return None
        # print(all_data)
        return all_data, not_found_items
    

class HardwareHelper():

    # sysbench版本架构不全，启用，使用简单python自行实现计算，memory读取速度影响较低，可以设定为固定值

    KEYWORDS_DICT = {
        'TEST_CPU_CMD' : ["sysbench", "cpu", "--cpu-max-prime=20000", "--time=1", "run"],
        'TEST_MEM_CMD' : ["sysbench", "memory", '--memory-block-size=1M', '--memory-total-size=500MB', 'run'],
        'TEST_IO_CMD' : ["sysbench", "fileio", "--file-total-size=100MB", "--file-num=1", "--file-test-mode=rndrw", "--time=1", "--max-requests=0", "--threads=1", "run"],
        # 'TEST_IO_CMD' : "sysbench fileio --file-total-size=100MB --file-num=1 --file-test-mode=rndrw --time=1 --max-requests=0 --threads=1 run | grep -E 'MiB/s'",

        'INSTALL_SYSBENCH_CMD' : ['sudo', 'apt', 'install', 'sysbench', '-y']

    }

    def __init__(self) -> None:
        # self._install_sysbench()
        # self._testCpuScore()
        # self._testMemScore()
        # self._testIOScore()
        pass
    
    def _install_sysbench(self):
            try:
                # 使用 dpkg -i 命令安装 .deb 包
                result = subprocess.run(self.KEYWORDS_DICT['INSTALL_SYSBENCH_CMD'], check=True, capture_output=True, text=True)
                # print(f"sysbench installed successfully!")
            except subprocess.CalledProcessError as e:
                logging.error(f"An error occurred while installing the sysbench: {e}")
            except Exception as e:
                logging.error(f"An unexpected error occurred: {e}")

    def _testCpuScore(self):
        def process_event(event):
            # 这里可以是你的事件处理逻辑
            pass
        
        def simulate_events(events_per_second):
            # 模拟事件生成器
            while True:
                time.sleep(1 / events_per_second)
                yield "event"

        def benchmark_cpu(events_per_second):
            events = simulate_events(events_per_second)
            start_time = time.time()
            events_processed = 0 
            while time.time() - start_time < 1:
                event = next(events)
                process_event(event)
                events_processed += 1
            return events_processed
        
        # speed = 500
        # try:
        #     # 运行命令
        #     results = subprocess.run(self.KEYWORDS_DICT['TEST_CPU_CMD'], check=True, capture_output=True, text=True)
        #     lines = results.stdout.split('\n')
        #     for line in lines:
        #         if 'events per second' in line:
        #             speed = _Utils_extract_numbers(line)
        # except subprocess.CalledProcessError as e:
        #     logging.error(f"An error occurred while run sysbench cpu test: {e}")
        # except Exception as e:
        #     logging.error(f"An unexpected error occurred when run sysbench cpu test: {e}")
        # # print('test cpu speed is:',speed)
        # return speed[0]
        result = 500
        try:
            result = benchmark_cpu(500)
        except Exception as e:
            logging.error(e)
        print(result)
        return result

    def _testMemScore(self):
        # speed = 16000
        # try:
        #     # 运行命令
        #     results = subprocess.run(self.KEYWORDS_DICT['TEST_MEM_CMD'], check=True, capture_output=True, text=True)
        #     lines = results.stdout.split('\n')
        #     for line in lines:
        #         if "MiB/s" in line:
        #             speed = _Utils_extract_numbers(line)
        # except subprocess.CalledProcessError as e:
        #     logging.error(f"An error occurred while run sysbench mem test: {e}")
        # except Exception as e:
        #     logging.error(f"An unexpected error occurred when run sysbench mem test: {e}")
        # # print('test mem speed is:',speed)
        # return speed[1]
        return 16000

    def _testIOScore(self):
        def test_write_speed(file_path, block_size=4096):
            """测试写入速度"""
            start_time = time.time()
            with open(file_path, 'wb') as file:
                for i in range(10000):  # 写入10000次
                    file.write(b'0' * block_size)  # 写入固定大小的数据
            end_time = time.time()
            write_time = end_time - start_time
            write_speed = block_size * 10000 / write_time  # 单位：字节/秒
            return write_speed
        
        def test_read_speed(file_path, block_size=4096):
            """测试读取速度"""
            file_size = os.path.getsize(file_path)
            with open(file_path, 'rb') as file:
                start_time = time.time()
                while file.read(block_size):  # 循环读取数据
                    pass
                end_time = time.time()
                read_time = end_time - start_time
                read_speed = file_size / read_time  # 单位：字节/秒
                return read_speed

        # speed = []
        # try:
        #     # 使用 dpkg -i 命令安装 .deb 包
        #     subprocess.run(["sysbench", "fileio", "--file-total-size=100MB", "--file-num=1", "prepare"],check=True, capture_output=True, text=True)
        #     results = subprocess.run(self.KEYWORDS_DICT['TEST_IO_CMD'], check=True, capture_output=True, text=True)
        #     lines = results.stdout.split('\n')
        #     for line in lines:
        #         if 'MiB/s' in line:
        #             speed.append(_Utils_extract_numbers(line))
        #     subprocess.run(["sysbench", "fileio", "--file-total-size=100MB", "--file-num=1", "cleanup"],check=True, capture_output=True, text=True)
        # except subprocess.CalledProcessError as e:
        #     logging.error(f"An error occurred while run sysbench mem test: {e}")
        # except Exception as e:
        #     logging.error(f"An unexpected error occurred when run sysbench mem test: {e}")
        # print('test io speed is:',speed)
        # return speed[0][0],speed[1][0]
        write_speed = test_write_speed("/var/cache/kylin-system-updater/IOtest.txt", 1024)/(1024 * 1024)
        read_speed = test_read_speed("/var/cache/kylin-system-updater/IOtest.txt", 1024)/(1024 * 1024)
        # print(read_speed, write_speed)
        return read_speed, write_speed

# deblist = ["apt","apache","systemd","linux-modules-extra-5.4.18-116-generic"]
# 读取文件内容
# with open("../installTest/installTestDebList.list", 'r') as file:
#     file_content = file.readlines()
# # 将内容拆分成包名称列表
# deblist = file_content[0].split()
# print(deblist)
# testclass = calTimeManager()
# print('预估计安装时间为：', testclass.CalInstallTimeByDebsList_min(deblist))
